import { CallbackType } from './awayback.model';
import { merge } from 'lodash';
/**
 * @license
 * awayback
 * Released under MIT license
 * Copyright Luca Raúl Joos
 */
function awayback() {
    const events = {};
    function create(event) {
        events[event] = {
            callbacks: [],
            data: [],
            runs: 0,
        };
    }
    function emit(event, ...data) {
        if (typeof events[event] === 'undefined')
            create(event);
        const self = events[event];
        if (!self)
            return;
        if (typeof data !== 'undefined') {
            self.data.push(data);
        }
        self.runs++;
        self.callbacks.forEach((callback) => {
            if (callback.type === CallbackType.on ||
                (callback.type === CallbackType.once && callback.runs === 0) ||
                (callback.type === CallbackType.only &&
                    callback.runs === 0 &&
                    self.callbacks.reduce((sum, callback) => sum + callback.runs, 0) === 0)) {
                callback.handler(...data);
                callback.runs++;
            }
        });
    }
    function listen(type, event, handler, options) {
        if (typeof events[event] === 'undefined')
            create(event);
        const self = events[event];
        if (!self)
            return;
        self.callbacks.push({
            handler: handler,
            type,
            runs: 0,
            options: merge({ isExecutingPrevious: false }, options !== null && options !== void 0 ? options : {}),
        });
        if (self.runs > 0) {
            self.callbacks.forEach((callback) => {
                var _a;
                if (!((_a = callback.options.isExecutingPrevious) !== null && _a !== void 0 ? _a : false))
                    return;
                let isExiting = false;
                while (callback.runs < self.runs && !isExiting) {
                    if (callback.type === CallbackType.on ||
                        (callback.type === CallbackType.once && callback.runs === 0) ||
                        (callback.type === CallbackType.only &&
                            callback.runs === 0 &&
                            self.callbacks.reduce((sum, callback) => sum + callback.runs, 0) === 0)) {
                        callback.handler(...self.data[callback.runs]);
                        callback.runs++;
                    }
                    else {
                        isExiting = true;
                    }
                }
            });
        }
    }
    function on(event, handler, options) {
        listen(CallbackType.on, event, handler, options);
    }
    function once(event, handler, options) {
        listen(CallbackType.once, event, handler, options);
    }
    function only(event, handler, options) {
        listen(CallbackType.only, event, handler, options);
    }
    function remove(event, handler) {
        if (typeof events[event] === 'undefined')
            return;
        const self = events[event];
        if (!self)
            return;
        self.callbacks = self.callbacks.filter((callback) => callback.handler !== handler);
    }
    return {
        events,
        emit,
        on,
        once,
        only,
        remove,
    };
}
export { CallbackType };
export default awayback;
